@@ -72,6 +72,10 @@ from itertools import chain, izip
 from twisted.internet import defer, protocol
 from twisted.protocols import policies
 
+try:
+    import hiredis
+except ImportError:
+    pass
 
 class RedisError(Exception):
     pass
@@ -93,7 +97,7 @@ class InvalidData(RedisError):
     pass
 
 
-class RedisBase(protocol.Protocol, policies.TimeoutMixin):
+class RedisBase(protocol.Protocol, policies.TimeoutMixin, object):
     """The main Redis client."""
 
     ERROR = "-"
@@ -104,7 +108,7 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
 
     def __init__(self, db=None, password=None, charset='utf8', errors='strict'):
         self.charset = charset
-        self.db = db
+        self.db = db if db is not None else 0
         self.password = password
         self.errors = errors
         self._buffer = ''
@@ -117,13 +121,12 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
     def dataReceived(self, data):
         """Receive data.
 
-        Spec: http://code.google.com/p/redis/wiki/ProtocolSpecification
-
+        Spec: http://redis.io/topics/protocol
         """
+        self.resetTimeout()
         self._buffer = self._buffer + data
 
         while self._buffer:
-            self.resetTimeout()
 
             # if we're expecting bulk data, read that many bytes
             if self._bulk_length is not None:
@@ -193,9 +196,22 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
 
     def connectionMade(self):
         """ Called when incoming connections is made to the server. """
-        self._disconnected = False
+        d = defer.succeed(True)
+
+        # if we have a password set, make sure we auth
         if self.password:
-            return self.auth(self.password)
+            d.addCallback(lambda _res : self.auth(self.password))
+
+        # select the db passsed in
+        if self.db:
+            d.addCallback(lambda _res : self.select(self.db))
+
+        def done_connecting(_res):
+            # set our state as soon as we're properly connected
+            self._disconnected = False
+        d.addCallback(done_connecting)
+
+        return d
 
     def connectionLost(self, reason):
         """Called when the connection is lost.
@@ -217,8 +233,13 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
 
     def errorReceived(self, data):
         """Error response received."""
-        reply = ResponseError(data[4:] if data[:4] == 'ERR ' else data)
-        self.responseReceived(reply)
+        reply = ResponseError(data if data[:4] == 'ERR ' else data)
+        if self._request_queue:
+            # properly errback this reply
+            self._request_queue.popleft().errback(reply)
+        else:
+            # we should have a request queue - if not, just raise this exception
+            raise reply
 
     def singleLineReceived(self, data):
         """Single line response received."""
@@ -251,20 +272,7 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
     def bulkDataReceived(self, data):
         """Bulk data response received."""
         self._bulk_length = None
-
-        # try to convert to int/float, otherwise treat as string
-        try:
-            if data is None:
-                element = None
-            elif '.' in data:
-                element = float(data)
-            else:
-                element = int(data)
-        except ValueError:
-            element = data
-            #element = data.decode(self.charset)
-
-        self.responseReceived(element)
+        self.responseReceived(data)
 
     def multiBulkDataReceived(self):
         """Multi bulk response received.
@@ -316,18 +324,18 @@ class RedisBase(protocol.Protocol, policies.TimeoutMixin):
                                   % (s.encode(self.charset, 'replace'), e))
         return str(s)
 
-    def _write(self, s):
-        """Send data."""
-        self.transport.write(s)
-
-    def _mb_cmd(self, *args):
-        """ Issue a multi-bulk command. """
+    def _send(self, *args):
+        """Encode and send a request using the 'unified request protocol' (aka multi-bulk)"""
         cmds = []
         for i in args:
             v = self._encode(i)
             cmds.append('$%s\r\n%s\r\n' % (len(v), v))
         cmd = '*%s\r\n' % len(args) + ''.join(cmds)
-        self._write(cmd)
+        self.transport.write(cmd)
+
+    def send(self, command, *args):
+        self._send(command, *args)
+        return self.getResponse()
 
 
 class Redis(RedisBase):
@@ -343,14 +351,46 @@ class Redis(RedisBase):
         """
         Test command. Expect PONG as a reply.
         """
-        self._write('PING\r\n')
+        self._send('PING')
+        return self.getResponse()
+
+    def shutdown(self):
+        """
+        Synchronously save the dataset to disk and then shut down the server
+        """
+        self._send('SHUTDOWN')
+        return self.getResponse()
+
+    def slaveof(self, host, port):
+        """
+        Make the server a slave of another instance, or promote it as master
+
+        The SLAVEOF command can change the replication settings of a slave on
+        the fly. If a Redis server is arleady acting as slave, the command
+        SLAVEOF NO ONE will turn off the replicaiton turning the Redis server
+        into a MASTER. In the proper form SLAVEOF hostname port will make the
+        server a slave of the specific server listening at the specified
+        hostname and port.
+
+        If a server is already a slave of some master, SLAVEOF hostname port
+        will stop the replication against the old server and start the
+        synchrnonization against the new one discarding the old dataset.
+
+        The form SLAVEOF no one will stop replication turning the server into a
+        MASTER but will not discard the replication. So if the old master stop
+        working it is possible to turn the slave into a master and set the
+        application to use the new master in read/write. Later when the other
+        Redis server will be fixed it can be configured in order to work as
+        slave.
+        """
+        self._send('SLAVEOF', host, port)
         return self.getResponse()
 
     def get_config(self, pattern):
         """
         Get configuration for Redis at runtime.
         """
-        self._mb_cmd('CONFIG', 'GET', pattern)
+        self._send('CONFIG', 'GET', pattern)
         def post_process(values):
             # transform into dict
             res = {}
@@ -365,7 +405,7 @@ class Redis(RedisBase):
         """
         Set configuration at runtime.
         """
-        self._mb_cmd('CONFIG', 'SET', parameter, value)
+        self._send('CONFIG', 'SET', parameter, value)
         return self.getResponse()
 
     # Commands operating on string values
@@ -384,11 +424,9 @@ class Redis(RedisBase):
             command = 'SET'
 
         if expire:
-            self._mb_cmd('SETEX', key, expire, value)
+            self._send('SETEX', key, expire, value)
         else:
-            value = self._encode(value)
-            self._write('%s %s %s\r\n%s\r\n'
-                        % (command, key, len(value), value))
+            self._send(command, key, value)
         return self.getResponse()
 
     def mset(self, mapping, preserve=False):
@@ -396,22 +434,22 @@ class Redis(RedisBase):
             command = 'MSETNX'
         else:
             command = 'MSET'
-        self._mb_cmd(command, *list(chain(*mapping.iteritems())))
+        self._send(command, *list(chain(*mapping.iteritems())))
         return self.getResponse()
 
     def append(self, key, value):
-        self._write('APPEND %s %s\r\n%s\r\n' % (
-                    key, len(value), value))
+        self._send('APPEND', key, value)
         return self.getResponse()
 
-    def substr(self, key, start, end):
-        self._mb_cmd('SUBSTR', key, start, end)
+    def getrange(self, key, start, end):
+        self._send('GETRANGE', key, start, end)
         return self.getResponse()
+    substr = getrange
 
     def get(self, key):
         """
         """
-        self._write('GET %s\r\n' % key)
+        self._send('GET', key)
         return self.getResponse()
 
     def getset(self, key, value):
@@ -422,43 +460,43 @@ class Redis(RedisBase):
     def mget(self, *args):
         """
         """
-        self._write('MGET %s\r\n' % ' '.join(args))
+        self._send('MGET', *args)
         return self.getResponse()
 
     def incr(self, key, amount=1):
         """
         """
         if amount == 1:
-            self._write('INCR %s\r\n' % key)
+            self._send('INCR', key)
         else:
-            self._write('INCRBY %s %s\r\n' % (key, amount))
+            self._send('INCRBY', key, amount)
         return self.getResponse()
 
     def decr(self, key, amount=1):
         """
         """
         if amount == 1:
-            self._write('DECR %s\r\n' % key)
+            self._send('DECR', key)
         else:
-            self._write('DECRBY %s %s\r\n' % (key, amount))
+            self._send('DECRBY', key, amount)
         return self.getResponse()
 
     def exists(self, key):
         """
         """
-        self._write('EXISTS %s\r\n' % key)
+        self._send('EXISTS', key)
         return self.getResponse()
 
     def delete(self, key):
         """
         """
-        self._write('DEL %s\r\n' % key)
+        self._send('DEL', key)
         return self.getResponse()
 
     def get_type(self, key):
         """
         """
-        self._write('TYPE %s\r\n' % key)
+        self._send('TYPE', key)
         res = self.getResponse()
         # return None if res == 'none' else res
         return res
@@ -467,7 +505,7 @@ class Redis(RedisBase):
     def keys(self, pattern):
         """
         """
-        self._write('KEYS %s\r\n' % pattern)
+        self._send('KEYS', pattern)
 
         def post_process(res):
             if res is not None:
@@ -482,54 +520,73 @@ class Redis(RedisBase):
         """
         """
         #raise NotImplementedError("Implemented but buggy, do not use.")
-        self._write('RANDOMKEY\r\n')
+        self._send('RANDOMKEY')
         return self.getResponse()
 
     def rename(self, src, dst, preserve=False):
         """
         """
-        if preserve:
-            self._write('RENAMENX %s %s\r\n' % (src, dst))
-            return self.getResponse()
-        else:
-            self._write('RENAME %s %s\r\n' % (src, dst))
-            return self.getResponse() #.strip()
+        self._send('RENAMENX' if preserve else 'RENAME', src, dst)
+        return self.getResponse() #.strip()
 
     def dbsize(self):
         """
+        Return the number of keys in the selected database
         """
-        self._write('DBSIZE\r\n')
+        self._send('DBSIZE')
         return self.getResponse()
 
     def expire(self, key, time):
         """
         """
-        self._write('EXPIRE %s %s\r\n' % (key, time))
+        self._send('EXPIRE', key, time)
         return self.getResponse()
 
     def ttl(self, key):
         """
         """
-        self._write('TTL %s\r\n' % key)
+        self._send('TTL', key)
         return self.getResponse()
 
+    # transaction commands:
     def multi(self):
-        self._write('MULTI\r\n')
+        """
+        Mark the start of a transaction block
+        """
+        self._send('MULTI')
         return self.getResponse()
 
     def execute(self):
-        """Sends the EXEC command
+        """
+        Sends the EXEC command
 
         Called execute because exec is a reserved word in Python.
-
         """
-        self._write('EXEC\r\n')
+        self._send('EXEC')
         return self.getResponse()
 
     def discard(self):
-        self._write('DISCARD\r\n')
+        """
+        Discard all commands issued after MULTI
+        """
+        self._send('DISCARD')
         return self.getResponse()
 
+    def watch(self, *keys):
+        """
+        Watch the given keys to determine execution of the MULTI/EXEC block
+        """
+        self._send('WATCH', *keys)
+        return self.getResponse()
+
+    def unwatch(self, *keys):
+        """
+        Forget about all watched keys
+        """
+        self._send('UNWATCH', *keys)
+        return self.getResponse()
+
+
     # # # # # # # # #
     # List Commands:
     # RPUSH
@@ -558,9 +615,17 @@ class Redis(RedisBase):
 
         @note Time complexity: O(1)
         """
-        value = self._encode(value)
-        self._write('%s %s %s\r\n%s\r\n'
-                    % ('RPUSH' if tail else 'LPUSH', key, len(value), value))
+        if tail:
+            return self.rpush(key, value)
+        else:
+            return self.lpush(key, value)
+
+    def lpush(self, key, value):
+        self._send('LPUSH', key, value)
+        return self.getResponse()
+
+    def rpush(self, key, value):
+        self._send('RPUSH', key, value)
         return self.getResponse()
 
     def llen(self, key):
@@ -574,7 +639,7 @@ class Redis(RedisBase):
 
         @note Time complexity: O(1)
         """
-        self._write('LLEN %s\r\n' % key)
+        self._send('LLEN', key)
         return self.getResponse()
 
     def lrange(self, key, start, end):
@@ -598,7 +663,7 @@ class Redis(RedisBase):
 
         @note Time complexity: O(n) (with n being the length of the range)
         """
-        self._write('LRANGE %s %s %s\r\n' % (key, start, end))
+        self._send('LRANGE', key, start, end)
         return self.getResponse()
 
     def ltrim(self, key, start, end):
@@ -623,7 +688,7 @@ class Redis(RedisBase):
 
         @note Time complexity: O(n) (with n being len of list - len of range)
         """
-        self._write('LTRIM %s %s %s\r\n' % (key, start, end))
+        self._send('LTRIM', key, start, end)
         return self.getResponse()
 
     def lindex(self, key, index):
@@ -642,7 +707,7 @@ class Redis(RedisBase):
         Note that even if the average time complexity is O(n) asking for
         the first or the last element of the list is O(1).
         """
-        self._write('LINDEX %s %s\r\n' % (key, index))
+        self._send('LINDEX', key, index)
         return self.getResponse()
 
     def pop(self, key, tail=False):
@@ -656,7 +721,7 @@ class Redis(RedisBase):
         If the key does not exist or the list is already empty the special
         value 'nil' is returned.
         """
-        self._write('%s %s\r\n' % ('RPOP' if tail else 'LPOP', key))
+        self._send('RPOP' if tail else 'LPOP', key)
         return self.getResponse()
 
     def bpop(self, keys, tail=False, timeout=30):
@@ -712,11 +777,7 @@ class Redis(RedisBase):
         client values will return false or nil accordingly to the
         programming language used.
         """
-        cmd = '%s ' % ('BRPOP' if tail else 'BLPOP')
-        for key in keys:
-            cmd += '%s ' % key
-        cmd += '%s\r\n' % str(timeout)
-        self._write(cmd)
+        self._send('BRPOP' if tail else 'BLPOP', *(list(keys) + [str(timeout)]))
         return self.getResponse()
 
     def rpoplpush(self, srckey, dstkey):
@@ -765,7 +826,7 @@ class Redis(RedisBase):
 
         Bulk reply
         """
-        self._write('%s %s %s\r\n' % ('RPOPLPUSH', srckey, dstkey))
+        self._send('RPOPLPUSH', srckey, dstkey)
         return self.getResponse()
 
     def lset(self, key, index, value):
@@ -784,9 +845,7 @@ class Redis(RedisBase):
 
         @note Time complexity: O(N) (with N being the length of the list)
         """
-        value = self._encode(value)
-        self._write('LSET %s %s %s\r\n%s\r\n'
-                    % (key, index, len(value), value))
+        self._send('LSET', key, index, value)
         return self.getResponse()
 
     def lrem(self, key, value, count=0):
@@ -810,144 +869,166 @@ class Redis(RedisBase):
 
         @note Time complexity: O(N) (with N being the length of the list)
         """
-        value = self._encode(value)
-        self._write('LREM %s %s %s\r\n%s\r\n'
-                    % (key, count, len(value), value))
+        self._send('LREM', key, count, value)
         return self.getResponse()
 
     # Commands operating on sets
+    def _list_to_set(self, res):
+        if type(res) is list:
+            return set(res)
+        return res
+
     def sadd(self, key, value):
         """
         """
-        value = self._encode(value)
-        self._write('SADD %s %s\r\n%s\r\n' % (key, len(value), value))
+        self._send('SADD', key, value)
         return self.getResponse()
 
     def srem(self, key, value):
         """
         """
-        value = self._encode(value)
-        self._write('SREM %s %s\r\n%s\r\n' % (key, len(value), value))
+        self._send('SREM', key, value)
         return self.getResponse()
 
     def spop(self, key):
-        self._write('SPOP %s\r\n' % key)
+        self._send('SPOP', key)
         return self.getResponse()
 
     def scard(self, key):
-        self._write('SCARD %s\r\n' % key)
+        self._send('SCARD', key)
         return self.getResponse()
 
     def sismember(self, key, value):
         """
         """
-        value = self._encode(value)
-        self._write('SISMEMBER %s %s\r\n%s\r\n' % (key, len(value), value))
+        self._send('SISMEMBER', key, value)
         return self.getResponse()
 
     def sdiff(self, *args):
-        self._mb_cmd('SDIFF', *args)
+        self._send('SDIFF', *args)
         return self.getResponse()
 
     def sdiffstore(self, dstkey, *args):
-        self._mb_cmd('SDIFFSTORE', dstkey, *args)
+        self._send('SDIFFSTORE', dstkey, *args)
         return self.getResponse()
 
     def srandmember(self, key):
-        self._mb_cmd('SRANDMEMBER', key)
+        self._send('SRANDMEMBER', key)
         return self.getResponse()
 
     def sinter(self, *args):
         """
         """
-        self._write('SINTER %s\r\n' % ' '.join(args))
-
-        def post_process(res):
-            if type(res) is list:
-                res = set(res)
-            return res
-
-        return self.getResponse().addCallback(post_process)
+        self._send('SINTER', *args)
+        return self.getResponse().addCallback(self._list_to_set)
 
     def sinterstore(self, dest, *args):
         """
         """
-        self._write('SINTERSTORE %s %s\r\n' % (dest, ' '.join(args)))
+        self._send('SINTERSTORE', dest, *args)
         return self.getResponse()
 
     def smembers(self, key):
         """
         """
-        self._write('SMEMBERS %s\r\n' % key)
-
-        def post_process(res):
-            if type(res) is list:
-                res = set(res)
-            return res
-
-        return self.getResponse().addCallback(post_process)
+        self._send('SMEMBERS', key)
+        return self.getResponse().addCallback(self._list_to_set)
 
     def smove(self, srckey, dstkey, member):
         """ Move the specifided member from the set at srckey to the set at dstkey. """
-        self._mb_cmd('SMOVE', srckey, dstkey, member)
+        self._send('SMOVE', srckey, dstkey, member)
         return self.getResponse()
 
     def sunion(self, *args):
         """
         """
-        self._write('SUNION %s\r\n' % ' '.join(args))
-
-        def post_process(res):
-            if type(res) is list:
-                res = set(res)
-            return res
-        return self.getResponse().addCallback(post_process)
+        self._send('SUNION', *args)
+        return self.getResponse().addCallback(self._list_to_set)
 
     def sunionstore(self, dest, *args):
         """
         """
-        self._write('SUNIONSTORE %s %s\r\n' % (dest, ' '.join(args)))
+        self._send('SUNIONSTORE', dest, *args)
         return self.getResponse()
 
     # Multiple databases handling commands
     def select(self, db):
         """
+        Select the DB with having the specified zero-based numeric index. New
+        connections always use DB 0.
         """
-        self._write('SELECT %s\r\n' % db)
+        self._send('SELECT', db)
         return self.getResponse()
 
     def move(self, key, db):
         """
         """
-        self._write('MOVE %s %s\r\n' % (key, db))
+        self._send('MOVE', key, db)
         return self.getResponse()
 
     def flush(self, all_dbs=False):
         """
         """
-        self._write('%s\r\n' % ('FLUSHALL' if all_dbs else 'FLUSHDB'))
+        if all_dbs:
+            return self.flushall()
+        else:
+            return self.flushdb()
+
+    def flushall(self):
+        self._send('FLUSHALL')
+        return self.getResponse()
+
+    def flushdb(self):
+        self._send('FLUSHDB')
         return self.getResponse()
 
     # Persistence control commands
+    def bgrewriteaof(self):
+        """
+        Rewrites the append-only file to reflect the current dataset in memory.
+        If BGREWRITEAOF fails, no data gets lost as the old AOF will be
+        untouched.
+        """
+        self._send('BGREWRITEAOF')
+        return self.getResponse()
+
+    def bgsave(self):
+        """
+        Save the DB in background. The OK code is immediately returned. Redis
+        forks, the parent continues to server the clients, the child saves the
+        DB on disk then exit. A client my be able to check if the operation
+        succeeded using the LASTSAVE command.
+        """
+        self._send('BGSAVE')
+        return self.getResponse()
+
     def save(self, background=False):
         """
+        Synchronously save the dataset to disk.
         """
         if background:
-            self._write('BGSAVE\r\n')
+            return self.bgsave()
         else:
-            self._write('SAVE\r\n')
+            self._send('SAVE')
         return self.getResponse()
 
     def lastsave(self):
         """
+        Return the UNIX TIME of the last DB save executed with success. A
+        client may check if a BGSAVE command succeeded reading the LASTSAVE
+        value, then issuing a BGSAVE command and checking at regular intervals
+        every N seconds if LASTSAVE changed.
         """
-        self._write('LASTSAVE\r\n')
+        self._send('LASTSAVE')
         return self.getResponse()
 
     def info(self):
         """
+        The info command returns different information and statistics about the
+        server in an format that's simple to parse by computers and easy to red
+        by huamns.
         """
-        self._write('INFO\r\n')
+        self._send('INFO')
 
         def post_process(res):
             info = dict()
@@ -967,27 +1048,56 @@ class Redis(RedisBase):
         """
         stmt = ['SORT', key]
         if by:
-            stmt.append("BY %s" % by)
+            stmt.extend(['BY', by])
         if start and num:
-            stmt.append("LIMIT %s %s" % (start, num))
+            stmt.extend(['LIMIT', start, num])
         if get is None:
             pass
         elif isinstance(get, basestring):
-            stmt.append("GET %s" % get)
+            stmt.extend(['GET', get])
         elif isinstance(get, list) or isinstance(get, tuple):
             for g in get:
-                stmt.append("GET %s" % g)
+                stmt.extend(['GET', get])
         else:
             raise RedisError("Invalid parameter 'get' for Redis sort")
         if desc:
             stmt.append("DESC")
         if alpha:
             stmt.append("ALPHA")
-        self._write(' '.join(stmt + ["\r\n"]))
+        self._send(*stmt)
         return self.getResponse()
 
     def auth(self, passwd):
-        self._write('AUTH %s\r\n' % passwd)
+        """
+        Request for authentication in a password protected Redis server. Redis
+        can be instructed to require a password before allowing clients to
+        execute commands. This is done using the requirepass directive in the
+        configuration file.  If password matches the password in the
+        configuration file, the server replies with the OK status code and
+        starts accepting commands. Otherwise, an error is returned and the
+        clients needs to try a new password.
+
+        Note: because of the high performance nature of Redis, it is possible
+        to try a lot of passwords in parallel in very short time, so make sure
+        to generate a strong and very long password so that this attack is
+        infeasible.
+        """
+        self._send('AUTH', passwd)
+        return self.getResponse()
+
+    def quit(self):
+        """
+        Ask the server to close the connection. The connection is closed as
+        soon as all pending replies have been written to the client.
+        """
+        self._send('QUIT')
+        return self.getResponse()
+
+    def echo(self, msg):
+        """
+        Returns message.
+        """
+        self._send('ECHO', msg)
         return self.getResponse()
 
     # # # # # # # # #
@@ -1003,22 +1113,44 @@ class Redis(RedisBase):
     # HVALS
     # HGETALL
     def hmset(self, key, in_dict):
+        """
+        Sets the specified fields to their respective values in the hash stored
+        at key. This command overwrites any existing fields in the hash. If key
+        does not exist, a new key holding a hash is created.
+        """
         fields = list(chain(*in_dict.iteritems()))
-        self._mb_cmd('HMSET', *([key] + fields))
+        self._send('HMSET', key, *fields)
         return self.getResponse()
 
     def hset(self, key, field, value, preserve=False):
+        """
+        Sets field in the hash stored at key to value. If key does not exist, a
+        new key holding a hash is created. If field already exists in the hash,
+        it is overwritten.
+        """
         if preserve:
-            self._mb_cmd('HSETNX', key, field, value)
+            return self.hsetnx(key, field, value)
         else:
-            self._mb_cmd('HSET', key, field, value)
+            self._send('HSET', key, field, value)
+            return self.getResponse()
+
+    def hsetnx(self, key, field, value):
+        """
+        Sets field in the hash stored at key to value, only if field does not
+        yet exist. If key does not exist, a new key holding a hash is created.
+        If field already exists, this operation has no effect.
+        """
+        self._send('HSETNX', key, field, value)
         return self.getResponse()
 
     def hget(self, key, field):
+        """
+        Returns the value associated with field in the hash stored at key.
+        """
         if isinstance(field, basestring):
-            self._mb_cmd('HGET', key, field)
+            self._send('HGET', key, field)
         else:
-            self._mb_cmd('HMGET', *([key] + field))
+            self._send('HMGET', *([key] + field))
 
         def post_process(values):
             if not values:
@@ -1030,32 +1162,61 @@ class Redis(RedisBase):
         return self.getResponse().addCallback(post_process)
     hmget = hget
 
+    def hget_value(self, key, field):
+        assert isinstance(field, basestring)
+        self._send('HGET', key, field)
+        return self.getResponse()
+
     def hkeys(self, key):
-        self._mb_cmd('HKEYS', key)
+        self._send('HKEYS', key)
         return self.getResponse()
 
     def hvals(self, key):
-        self._mb_cmd('HVALS', key)
+        self._send('HVALS', key)
         return self.getResponse()
 
     def hincr(self, key, field, amount=1):
-        self._mb_cmd('HINCRBY', key, field, amount)
+        """
+        Increments the number stored at field in the hash stored at key by
+        increment. If key does not exist, a new key holding a hash is created.
+        If field does not exist or holds a string that cannot be interpreted as
+        integer, the value is set to 0 before the operation is performed.  The
+        range of values supported by HINCRBY is limited to 64 bit signed
+        integers.
+        """
+        self._send('HINCRBY', key, field, amount)
         return self.getResponse()
+    hincrby = hincr
 
     def hexists(self, key, field):
-        self._mb_cmd('HEXISTS', key, field)
+        """
+        Returns if field is an existing field in the hash stored at key.
+        """
+        self._send('HEXISTS', key, field)
         return self.getResponse()
 
-    def hdelete(self, key, field):
-        self._mb_cmd('HDEL', key, field)
+    def hdel(self, key, field):
+        """
+        Removes field from the hash stored at key.
+        """
+        self._send('HDEL', key, field)
         return self.getResponse()
+    hdelete = hdel # backwards compat for older txredis
 
     def hlen(self, key):
-        self._mb_cmd('HLEN', key)
+        """
+        Returns the number of fields contained in the hash stored at key.
+        """
+        self._send('HLEN', key)
         return self.getResponse()
 
     def hgetall(self, key):
-        self._mb_cmd('HGETALL', key)
+        """
+        Returns all fields and values of the hash stored at key. In the
+        returned value, every field name is followed by its value, so the
+        length of the reply is twice the size of the hash.
+        """
+        self._send('HGETALL', key)
 
         def post_process(key_vals):
             res = {}
@@ -1071,8 +1232,7 @@ class Redis(RedisBase):
         """
         Publishes a message to all subscribers of a specified channel.
         """
-        self._write('PUBLISH %s %s\r\n%s\r\n'
-                    % (channel, len(message), message))
+        self._send('PUBLISH', channel, message)
         return self.getResponse()
 
     # # # # # # # # #
@@ -1091,19 +1251,19 @@ class Redis(RedisBase):
     # ZREMRANGEBYSCORE
     # ZUNIONSTORE / ZINTERSTORE
     def zadd(self, key, member, score):
-        self._mb_cmd('ZADD', key, score, member)
+        self._send('ZADD', key, score, member)
         return self.getResponse()
 
     def zrem(self, key, member):
-        self._mb_cmd('ZREM', key, member)
+        self._send('ZREM', key, member)
         return self.getResponse()
 
     def zremrangebyrank(self, key, start, end):
-        self._mb_cmd('ZREMRANGEBYRANK', key, start, end)
+        self._send('ZREMRANGEBYRANK', key, start, end)
         return self.getResponse()
 
     def zremrangebyscore(self, key, min, max):
-        self._mb_cmd('ZREMRANGEBYSCORE', key, min, max)
+        self._send('ZREMRANGEBYSCORE', key, min, max)
         return self.getResponse()
 
     def _zopstore(self, op, dstkey, keys, aggregate=None):
@@ -1124,7 +1284,7 @@ class Redis(RedisBase):
         if aggregate:
             args.append('AGGREGATE')
             args.append(aggregate)
-        self._mb_cmd(*args)
+        self._send(*args)
         return self.getResponse()
 
     def zunionstore(self, dstkey, keys, aggregate=None):
@@ -1142,16 +1302,16 @@ class Redis(RedisBase):
         return self._zopstore('ZINTERSTORE', dstkey, keys, aggregate)
 
     def zincr(self, key, member, incr=1):
-        self._mb_cmd('ZINCRBY', key, incr, member)
+        self._send('ZINCRBY', key, incr, member)
         return self.getResponse()
 
     def zrank(self, key, member, reverse=False):
         cmd = 'ZREVRANK' if reverse else 'ZRANK'
-        self._mb_cmd(cmd, key, member)
+        self._send(cmd, key, member)
         return self.getResponse()
 
     def zcount(self, key, min, max):
-        self._mb_cmd('ZCOUNT', key, min, max)
+        self._send('ZCOUNT', key, min, max)
         return self.getResponse()
 
     def zrange(self, key, start, end, withscores=False, reverse=False):
@@ -1159,7 +1319,7 @@ class Redis(RedisBase):
         args = [cmd, key, start, end]
         if withscores:
             args.append('WITHSCORES')
-        self._mb_cmd(*args)
+        self._send(*args)
         dfr = self.getResponse()
 
         def post_process(vals_and_scores):
@@ -1168,7 +1328,7 @@ class Redis(RedisBase):
             bins = len(vals_and_scores) - 1
             i = 0
             while i < bins:
-                res.append((vals_and_scores[i], vals_and_scores[i+1]))
+                res.append((vals_and_scores[i], float(vals_and_scores[i+1])))
                 i += 2
             return res
 
@@ -1179,13 +1339,22 @@ class Redis(RedisBase):
     def zrevrange(self, key, start, end, withscores=False):
         return self.zrange(key, start, end, withscores, reverse=True)
 
+    def zrevrank(self, key, member):
+        self._send('ZREVRANK', key, member)
+        return self.getResponse()
+
     def zcard(self, key):
-        self._write('ZCARD %s\r\n' % key)
+        self._send('ZCARD', key)
         return self.getResponse()
 
     def zscore(self, key, element):
-        self._mb_cmd('ZSCORE', key, element)
-        return self.getResponse()
+        self._send('ZSCORE', key, element)
+        def post_process(res):
+            if res is not None:
+                return float(res)
+            else:
+                return res
+        return self.getResponse().addCallback(post_process)
 
     def zrangebyscore(self, key, min='-inf', max='+inf', offset=None,
                       count=None, withscores=False):
@@ -1194,7 +1363,7 @@ class Redis(RedisBase):
             args.extend(['LIMIT', offset, count])
         if withscores:
             args.append('WITHSCORES')
-        self._mb_cmd(*args)
+        self._send(*args)
         dfr = self.getResponse()
 
         def post_process(vals_and_scores):
@@ -1203,7 +1372,7 @@ class Redis(RedisBase):
             bins = len(vals_and_scores) - 1
             i = 0
             while i < bins:
-                res.append((vals_and_scores[i], vals_and_scores[i+1]))
+                res.append((vals_and_scores[i], float(vals_and_scores[i+1])))
                 i += 2
             return res
 
@@ -1211,6 +1380,29 @@ class Redis(RedisBase):
             dfr.addCallback(post_process)
         return dfr
 
+class HiRedisProtocol(Redis):
+    """ A subclass of the Redis protocol that uses the hiredis library for parsing. """
+    def __init__(self, db=None, password=None, charset='utf8', errors='strict'):
+        Redis.__init__(self, db, password, charset, errors)
+        self._reader = hiredis.Reader(protocolError=InvalidData,
+                                      replyError=ResponseError)
+
+    def dataReceived(self, data):
+        """Receive data.
+        """
+        self.resetTimeout()
+        if data:
+            self._reader.feed(data)
+        res = self._reader.gets()
+        while res is not False:
+            if isinstance(res, ResponseError):
+                self._request_queue.popleft().errback(res)
+            else:
+                if isinstance(res, basestring) and res == 'none':
+                    res = None
+                self._request_queue.popleft().callback(res)
+            res = self._reader.gets()
+
 
 class RedisSubscriber(RedisBase):
     """
@@ -1291,7 +1483,7 @@ class RedisSubscriber(RedisBase):
         this connection, then channelSubscribed will not be invoked but the
         channel will continue to be subscribed to.
         """
-        self._write("SUBSCRIBE %s\r\n" % ' '.join(channels))
+        self._send('SUBSCRIBE', *channels)
 
     def unsubscribe(self, *channels):
         """
@@ -1302,10 +1494,7 @@ class RedisSubscriber(RedisBase):
         provided that is not subscribed to by this connection, then
         channelUnsubscribed will not be invoked.
         """
-        if len(channels) == 0:
-            self._write("UNSUBSCRIBE\r\n")
-        else:
-            self._write("UNSUBSCRIBE %s\r\n" % ' '.join(channels))
+        self._send('UNSUBSCRIBE', *channels)
 
     def psubscribe(self, *patterns):
         """
@@ -1315,7 +1504,7 @@ class RedisSubscriber(RedisBase):
         channelPatternSubscribed will be invoked for each channel pattern that
         is subscribed to.
         """
-        self._write("PSUBSCRIBE %s\r\n" % ' '.join(patterns))
+        self._send('PSUBSCRIBE', *patterns)
 
     def punsubscribe(self, *patterns):
         """
@@ -1325,10 +1514,7 @@ class RedisSubscriber(RedisBase):
         channelPatternUnsubscribed will be invoked for eeach channel pattern
         that is unsubscribed from.
         """
-        if len(patterns) == 0:
-            self._write("PUNSUBSCRIBE\r\n")
-        else:
-            self._write("PUNSUBSCRIBE %s\r\n" % ' '.join(patterns))
+        self._send('PUNSUBSCRIBE', *patterns)
 
 
 class RedisClientFactory(protocol.ReconnectingClientFactory):
