@@ -5,7 +5,7 @@ from twisted.internet import error
 from twisted.internet import protocol
 from twisted.internet import reactor
 from twisted.internet import defer
-from twisted.internet.task import Clock, deferLater
+from twisted.internet.task import Clock
 from twisted.test.proto_helpers import StringTransportWithDisconnection
 from twisted.trial import unittest
 
@@ -13,15 +13,25 @@ from txredis.protocol import Redis, RedisSubscriber, RedisClientFactory
 from txredis.protocol import ResponseError
 
 REDIS_HOST = 'localhost'
-REDIS_PORT = 6379
+REDIS_PORT = 6381
 
 
 class CommandsTestBase(unittest.TestCase):
-
-    @defer.inlineCallbacks
+    protocol = Redis
     def setUp(self):
-        clientCreator = protocol.ClientCreator(reactor, Redis)
-        self.redis = yield clientCreator.connectTCP(REDIS_HOST, REDIS_PORT)
+        clientCreator = protocol.ClientCreator(reactor, self.protocol)
+        d = clientCreator.connectTCP(REDIS_HOST, REDIS_PORT)
+        def got_conn(redis):
+            self.redis = redis
+        d.addCallback(got_conn)
+        def cannot_conn(res):
+            msg = '\n' * 3 + '*' * 80 + '\n' * 2
+            msg += "NOTE: Redis server not running on port %s. Please start a local instance of Redis " \
+                  "on this port to run unit tests against.\n\n" % REDIS_PORT
+            msg += '*' * 80 + '\n' * 4
+            raise unittest.SkipTest(msg)
+        d.addErrback(cannot_conn)
+        return d
 
     def tearDown(self):
         self.redis.transport.loseConnection()
@@ -37,7 +47,6 @@ class General(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_config(self):
-        r = self.redis
         t = self.assertEqual
         a = yield self.redis.get_config('*')
         self.assertTrue(isinstance(a, dict))
@@ -124,6 +133,7 @@ class General(CommandsTestBase):
         a = yield r.get_type('zzz')
         ex = None
         t(a, ex)
+        self.assertTrue(a == None or a == 'none')
 
     @defer.inlineCallbacks
     def test_keys(self):
@@ -164,20 +174,27 @@ class General(CommandsTestBase):
         ex = True
         t(a, ex)
 
+    def test_rename_same_src_dest(self):
+        r = self.redis
+        t = self.assertEqual
+        d = r.rename('a', 'a')
+        self.failUnlessFailure(d, ResponseError)
+        def test_err(a):
+            ex = ResponseError('ERR source and destination objects are the same')
+            t(str(a), str(ex))
+        d.addCallback(test_err)
+        return d
+
     @defer.inlineCallbacks
     def test_rename(self):
         r = self.redis
         t = self.assertEqual
 
-        a = yield r.rename('a', 'a')
-        ex = ResponseError('source and destination objects are the same')
-        t(str(a), str(ex))
         a = yield r.rename('a', 'b')
         ex = 'OK'
         t(a, ex)
-        a = yield r.rename('a', 'b')
-        ex = ResponseError('no such key')
-        t(str(a), str(ex))
+        a = yield r.get('a')
+        t(a, None)
         a = yield r.set('a', 1)
         ex = 'OK'
         t(a, ex)
@@ -188,11 +205,9 @@ class General(CommandsTestBase):
     @defer.inlineCallbacks
     def test_dbsize(self):
         r = self.redis
-        t = self.assertEqual
-
-        a = type((yield r.dbsize()))
-        ex = int
-        t(a, ex)
+        t = self.assertTrue
+        a = yield r.dbsize()
+        t(isinstance(a, int) or isinstance(a, long))
 
     @defer.inlineCallbacks
     def test_expire(self):
@@ -217,8 +232,16 @@ class General(CommandsTestBase):
         a = yield r.set('q', 1, expire=10)
         ex = 'OK'
         t(a, ex)
+        # the following checks the expected response of an EXPIRE on a key with
+        # an existing TTL. unfortunately the behaviour of redis changed in
+        # v2.1.3 so we have to determine which behaviour to expect...
+        info = yield r.info()
+        redis_vern = tuple(map(int, info['redis_version'].split('.')))
+        if redis_vern < (2, 1, 3):
+            ex = 0
+        else:
+            ex = 1
         a = yield r.expire('q', 1)
-        ex = 0
         t(a, ex)
 
     @defer.inlineCallbacks
@@ -241,7 +264,10 @@ class General(CommandsTestBase):
 
         string = 'This is a string'
         r.set('s', string)
-        a = yield r.substr('s', 0, 3)
+        a = yield r.substr('s', 0, 3) # old name
+        ex = 'This'
+        t(a, ex)
+        a = yield r.getrange('s', 0, 3) # new name
         ex = 'This'
         t(a, ex)
 
@@ -338,34 +364,30 @@ class General(CommandsTestBase):
         ex = 'OK'
         t(a, ex)
 
-    @defer.inlineCallbacks
-    def test_save(self):
-        r = self.redis
-        t = self.assertEqual
-
-        a = yield r.save()
-        ex = 'OK'
-        t(a, ex)
-        resp = yield r.save(background=True)
-        """
-        ex = ResponseError(
-        ...     assert str(e) == 'background save already in progress', str(e)
-        ... else:
-        ...     assert resp == 'OK'
-        """
-
-    @defer.inlineCallbacks
     def test_lastsave(self):
         r = self.redis
         t = self.assertEqual
 
         tme = int(time.time())
-        a = yield r.save()
-        ex = 'OK'
-        t(a, ex)
-        a = (yield r.lastsave()) >= tme
-        ex = True
-        t(a, ex)
+        d = r.save()
+        def done_save(a):
+            ex = 'OK'
+            t(a, ex)
+            d = r.lastsave()
+            def got_lastsave(a):
+                a = a >= tme
+                ex = True
+                t(a, ex)
+            d.addCallback(got_lastsave)
+            return d
+
+        def save_err(res):
+            if 'Background save already in progress' in str(res):
+                return True
+            else:
+                raise res
+        d.addCallbacks(done_save, save_err)
+        return d
 
     @defer.inlineCallbacks
     def test_info(self):
@@ -387,10 +409,6 @@ class General(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_execute(self):
-        # exec without multi will return ResponseError
-        r = yield self.redis.execute()
-        self.assertEqual(str(r), 'EXEC without MULTI')
-
         # multi with two sets
         yield self.redis.multi()
         r = yield self.redis.set('foo', 'bar')
@@ -402,21 +420,39 @@ class General(CommandsTestBase):
         r = yield self.redis.get('foo')
         self.assertEqual(r, 'barbar')
 
-    @defer.inlineCallbacks
     def test_discard(self):
+        d = self.redis.execute()
         # discard without multi will return ResponseError
-        r = yield self.redis.execute()
-        self.assertEqual(str(r), 'EXEC without MULTI')
+        d = self.failUnlessFailure(d, ResponseError)
 
         # multi with two sets
-        yield self.redis.set('foo', 'bar1')
-        yield self.redis.multi()
-        r = yield self.redis.set('foo', 'bar2')
-        r = yield self.redis.discard()
-        self.assertEqual(r, 'OK')
-        r = yield self.redis.get('foo')
-        self.assertEqual(r, 'bar1')
+        def step1(_res):
+            d = self.redis.set('foo', 'bar1')
 
+            def step2(_res):
+                d = self.redis.multi()
+                def in_multi(_res):
+                    d = self.redis.set('foo', 'bar2')
+                    def step3(_res):
+                        d = self.redis.discard()
+                        def step4(r):
+                            self.assertEqual(r, 'OK')
+                            d = self.redis.get('foo')
+                            def got_it(res):
+                                self.assertEqual(res, 'bar1')
+                            d.addCallback(got_it)
+                            return d
+                        d.addCallback(step4)
+                        return d
+                    d.addCallback(step3)
+                    return d
+                d.addCallback(in_multi)
+                return d
+            d.addCallback(step2)
+            return d
+
+        d.addCallback(step1)
+        return d
 
 class Strings(CommandsTestBase):
     """Test commands that operate on string values.
@@ -448,7 +484,7 @@ class Strings(CommandsTestBase):
         self.assertEqual(a, 0)
 
         a = yield self.redis.get('b')
-        self.assertEqual(a, 105.2)
+        self.assertEqual(a, '105.2')
 
     @defer.inlineCallbacks
     def test_get(self):
@@ -468,7 +504,7 @@ class Strings(CommandsTestBase):
         t(a, u'pippo')
 
         a = yield r.get('b')
-        ex = 15
+        ex = '15'
         t(a, ex)
 
         a = yield r.get('d')
@@ -476,7 +512,7 @@ class Strings(CommandsTestBase):
         t(a, ex)
 
         a = yield r.get('b')
-        ex = 15
+        ex = '15'
         t(a, ex)
 
         a = yield r.get('c')
@@ -518,7 +554,7 @@ class Strings(CommandsTestBase):
         ex = 'OK'
         t(a, ex)
         a = yield r.mget('a', 'b', 'c', 'd')
-        ex = [u'pippo', 15,
+        ex = [u'pippo', '15',
               u'\\r\\naaa\\nbbb\\r\\ncccc\\nddd\\r\\n', u'\\r\\n']
         t(a, ex)
 
@@ -613,10 +649,6 @@ class Lists(CommandsTestBase):
         a = yield r.set('a', 'a')
         ex = 'OK'
         t(a, ex)
-        a = yield r.push('a', 'a')
-        ex = ResponseError('Operation against a key holding the wrong kind '+
-                           'of value')
-        t(str(a), str(ex))
 
     @defer.inlineCallbacks
     def test_llen(self):
@@ -759,21 +791,53 @@ class Lists(CommandsTestBase):
         ex = None
         t(a, ex)
 
+    def test_lset_on_nonexistant_key(self):
+        r = self.redis
+        t = self.assertEqual
+
+        d = r.delete('l')
+        def bad_lset(_res):
+            d = r.lset('l', 0, 'a')
+            self.failUnlessFailure(d, ResponseError)
+            def match_err(a):
+                ex = ResponseError('ERR no such key')
+                t(str(a), str(ex))
+            d.addCallback(match_err)
+            return d
+        d.addCallback(bad_lset)
+        return d
+
+    def test_lset_bad_range(self):
+        r = self.redis
+        t = self.assertEqual
+
+        d = r.delete('l')
+        def proceed(_res):
+            d = r.push('l', 'aaa')
+            def done_push(a):
+                ex = 1
+                t(a, ex)
+                d = r.lset('l', 1, 'a')
+                self.failUnlessFailure(d, ResponseError)
+                def check(a):
+                    ex = ResponseError('ERR index out of range')
+                    t(str(a), str(ex))
+                d.addCallback(check)
+                return d
+            d.addCallback(done_push)
+            return d
+        d.addCallback(proceed)
+        return d
+
     @defer.inlineCallbacks
     def test_lset(self):
         r = self.redis
         t = self.assertEqual
 
         yield r.delete('l')
-        a = yield r.lset('l', 0, 'a')
-        ex = ResponseError('no such key')
-        t(str(a), str(ex))
         a = yield r.push('l', 'aaa')
         ex = 1
         t(a, ex)
-        a = yield r.lset('l', 1, 'a')
-        ex = ResponseError('index out of range')
-        t(str(a), str(ex))
         a = yield r.lset('l', 0, 'bbb')
         ex = 'OK'
         t(a, ex)
@@ -983,13 +1047,6 @@ class Sets(CommandsTestBase):
         a = yield r.sadd('s3', 'b')
         ex = 1
         t(a, ex)
-        a = yield r.sinter()
-        ex = ResponseError("wrong number of arguments for 'sinter' command")
-        t(str(a), str(ex))
-        a = yield r.sinter('l')
-        ex = ResponseError('Operation against a key holding the wrong kind '+
-                           'of value')
-        t(str(a), str(ex))
         a = yield r.sinter('s1', 's2', 's3')
         ex = set([])
         t(a, ex)
@@ -1038,10 +1095,6 @@ class Sets(CommandsTestBase):
         a = yield r.sadd('s', 'b')
         ex = 1
         t(a, ex)
-        a = yield r.smembers('l')
-        ex = ResponseError('Operation against a key holding the wrong kind '+
-                           'of value')
-        t(str(a), str(ex))
         a = yield r.smembers('s')
         ex = set([u'a', u'b'])
         t(a, ex)
@@ -1097,10 +1150,28 @@ class Sets(CommandsTestBase):
         ex = set([u'a', u'b'])
         t(a, ex)
 
+    @defer.inlineCallbacks
+    def test_sort_style(self):
+        # considering, given that redis only stores strings, whether the sort it
+        # provides is a numeric or a lexicographical sort; turns out that it's
+        # numeric; i.e. redis is doing implicit type coercion for the sort of
+        # numeric values.  This test serves to document that, and to a lesser
+        # extent check for regression in the implicit str() marshalling of txredis
+        r = self.redis
+        t = self.assertEqual
+        yield r.delete('l')
+        items = [ 007, 10, -5, 0.1, 100, -3, 20, 0.02, -3.141 ]
+        for i in items:
+            yield r.push('l', i, tail=True)
+        a = yield r.sort('l')
+        ex = map(str, sorted(items))
+        t(a, ex)
+
     @defer.inlineCallbacks
     def test_sort(self):
         r = self.redis
         t = self.assertEqual
+        s = lambda l: map(str, l)
 
         yield r.delete('l')
         a = yield r.push('l', 'ccc')
@@ -1124,27 +1195,27 @@ class Sets(CommandsTestBase):
         for i in range(1, 5):
             yield r.push('l', 1.0 / i, tail=True)
         a = yield r.sort('l')
-        ex = [0.25, 0.333333333333, 0.5, 1.0]
+        ex = s([0.25, 0.333333333333, 0.5, 1.0])
         t(a, ex)
         a = yield r.sort('l', desc=True)
-        ex = [1.0, 0.5, 0.333333333333, 0.25]
+        ex = s([1.0, 0.5, 0.333333333333, 0.25])
         t(a, ex)
         a = yield r.sort('l', desc=True, start=2, num=1)
-        ex = [0.333333333333]
+        ex = s([0.333333333333])
         t(a, ex)
         a = yield r.set('weight_0.5', 10)
         ex = 'OK'
         t(a, ex)
         a = yield r.sort('l', desc=True, by='weight_*')
-        ex = [0.5, 1.0, 0.333333333333, 0.25]
+        ex = s([0.5, 1.0, 0.333333333333, 0.25])
         t(a, ex)
         for i in (yield r.sort('l', desc=True)):
             yield r.set('test_%s' % i, 100 - float(i))
         a = yield r.sort('l', desc=True, get='test_*')
-        ex = [99.0, 99.5, 99.6666666667, 99.75]
+        ex = s([99.0, 99.5, 99.6666666667, 99.75])
         t(a, ex)
         a = yield r.sort('l', desc=True, by='weight_*', get='test_*')
-        ex = [99.5, 99.0, 99.6666666667, 99.75]
+        ex = s([99.5, 99.0, 99.6666666667, 99.75])
         t(a, ex)
         a = yield r.sort('l', desc=True, by='weight_*', get='missing_*')
         ex = [None, None, None, None]
@@ -1163,7 +1234,7 @@ class Sets(CommandsTestBase):
             a = yield r.set(key, value)
             t('OK', a)
             rval = yield r.get(key)
-            t(value, rval)
+            t(rval, str(value))
 
 
 class Hash(CommandsTestBase):
@@ -1179,6 +1250,24 @@ class Hash(CommandsTestBase):
         a = yield self.redis.hgetall('h')
         self.assertEquals(a, {'blank': ''})
 
+    @defer.inlineCallbacks
+    def test_cas(self):
+        r = self.redis
+        t = self.assertEqual
+
+        a = yield r.delete('h')
+        ex = 1
+        t(a, ex)
+
+        a = yield r.hsetnx('h', 'f', 'v')
+        ex = 1
+        t(a, ex)
+
+        a = yield r.hsetnx('h', 'f', 'v')
+        ex = 0
+        t(a, ex)
+
+
     @defer.inlineCallbacks
     def test_basic(self):
         r = self.redis
@@ -1316,7 +1405,7 @@ class LargeMultiBulk(CommandsTestBase):
         for i in data:
             r.sadd('s', i)
         res = yield r.smembers('s')
-        t(res, data)
+        t(res, set(map(str, data)))
 
 
 class SortedSet(CommandsTestBase):
@@ -1401,6 +1490,10 @@ class SortedSet(CommandsTestBase):
         t = self.assertEqual
 
         yield r.delete('z')
+        a = yield r.zrangebyscore('z', -1, -1, withscores=True)
+        ex = []
+        t(a, ex)
+
         yield r.zadd('z', 'a', 1.014)
         yield r.zadd('z', 'b', 4.252)
         yield r.zadd('z', 'c', 0.232)
@@ -1420,6 +1513,20 @@ class SortedSet(CommandsTestBase):
         a = yield r.zrangebyscore('z', min=1, offset=1, count=2, withscores=True)
         ex = [('b', 4.252), ('d', 10.425)]
 
+    @defer.inlineCallbacks
+    def test_zscore_and_zrange_nonexistant(self):
+        r = self.redis
+        t = self.assertEqual
+
+        yield r.delete('a')
+        a = yield r.zscore('a', 'somekey')
+        t(a, None)
+
+        yield r.delete('a')
+        a = yield r.zrange('a', 0, -1, withscores=True)
+        t(a, [])
+
+
     @defer.inlineCallbacks
     def test_zaggregatestore(self):
         r = self.redis
@@ -1526,6 +1633,27 @@ class BlockingListOperartions(CommandsTestBase):
         r2.transport.loseConnection()
 
 
+# if hiredis and its python wrappers are installed, test them too
+try:
+    import hiredis
+    from txredis.protocol import HiRedisProtocol
+    class HiRedisGeneral(General):
+        protcol = HiRedisProtocol
+    class HiRedisStrings(Strings):
+        protocol = HiRedisProtocol
+    class HiRedisLists(Lists):
+        protocol = HiRedisProtocol
+    class HiRedisHash(Hash):
+        protocol = HiRedisProtocol
+    class HiRedisSortedSet(SortedSet):
+        protocol = HiRedisProtocol
+    class HiRedisSets(Sets):
+        protocol = HiRedisProtocol
+    _hush_pyflakes = hiredis
+    del _hush_pyflakes
+except ImportError:
+    pass
+
 class Network(unittest.TestCase):
 
     def setUp(self):
@@ -1565,6 +1693,7 @@ class Network(unittest.TestCase):
         return done.addCallback(checkFailures)
 
 
+
 class Protocol(unittest.TestCase):
 
     def setUp(self):
@@ -1576,20 +1705,21 @@ class Protocol(unittest.TestCase):
     def sendResponse(self, data):
         self.proto.dataReceived(data)
 
-    @defer.inlineCallbacks
     def test_error_response(self):
         # pretending 'foo' is a set, so get is incorrect
         d = self.proto.get("foo")
-        self.assertEquals(self.transport.value(), "GET foo\r\n")
+        self.assertEquals(self.transport.value(), '*2\r\n$3\r\nGET\r\n$3\r\nfoo\r\n')
         msg = "Operation against a key holding the wrong kind of value"
         self.sendResponse("-%s\r\n" % msg)
-        r = yield d
-        self.assertEquals(str(r), msg)
+        self.failUnlessFailure(d, ResponseError)
+        def check_err(r):
+            self.assertEquals(str(r), msg)
+        return d
 
     @defer.inlineCallbacks
     def test_singleline_response(self):
         d = self.proto.ping()
-        self.assertEquals(self.transport.value(), "PING\r\n")
+        self.assertEquals(self.transport.value(), '*1\r\n$4\r\nPING\r\n')
         self.sendResponse("+PONG\r\n")
         r = yield d
         self.assertEquals(r, 'PONG')
@@ -1597,7 +1727,7 @@ class Protocol(unittest.TestCase):
     @defer.inlineCallbacks
     def test_bulk_response(self):
         d = self.proto.get("foo")
-        self.assertEquals(self.transport.value(), "GET foo\r\n")
+        self.assertEquals(self.transport.value(), '*2\r\n$3\r\nGET\r\n$3\r\nfoo\r\n')
         self.sendResponse("$3\r\nbar\r\n")
         r = yield d
         self.assertEquals(r, 'bar')
@@ -1605,7 +1735,8 @@ class Protocol(unittest.TestCase):
     @defer.inlineCallbacks
     def test_multibulk_response(self):
         d = self.proto.lrange("foo", 0, 1)
-        self.assertEquals(self.transport.value(), "LRANGE foo 0 1\r\n")
+        expected = '*4\r\n$6\r\nLRANGE\r\n$3\r\nfoo\r\n$1\r\n0\r\n$1\r\n1\r\n'
+        self.assertEquals(self.transport.value(), expected)
         self.sendResponse("*2\r\n$3\r\nbar\r\n$6\r\nlolwut\r\n")
         r = yield d
         self.assertEquals(r, ['bar', 'lolwut'])
@@ -1613,22 +1744,32 @@ class Protocol(unittest.TestCase):
     @defer.inlineCallbacks
     def test_integer_response(self):
         d = self.proto.dbsize()
-        self.assertEquals(self.transport.value(), "DBSIZE\r\n")
+        self.assertEquals(self.transport.value(), '*1\r\n$6\r\nDBSIZE\r\n')
         self.sendResponse(":1234\r\n")
         r = yield d
         self.assertEquals(r, 1234)
 
-
-class TestFactory(unittest.TestCase):
-    @defer.inlineCallbacks
+class TestFactory(CommandsTestBase):
     def setUp(self):
-        self.factory = RedisClientFactory()
-        reactor.connectTCP(REDIS_HOST, REDIS_PORT, self.factory)
-        yield self.factory.deferred
+        d = CommandsTestBase.setUp(self)
+        def do_setup(_res):
+            self.factory = RedisClientFactory()
+            reactor.connectTCP(REDIS_HOST, REDIS_PORT, self.factory)
+            d = self.factory.deferred
+            def cannot_connect(_res):
+                raise unittest.SkipTest('Cannot connect to Redis.')
+            d.addErrback(cannot_connect)
+            return d
+        d.addCallback(do_setup)
+        return d
 
     def tearDown(self):
+        CommandsTestBase.tearDown(self)
+        self.factory.continueTrying = 0
         self.factory.stopTrying()
-        self.factory.client.transport.loseConnection()
+        if self.factory.client:
+            self.factory.client.setTimeout(None)
+            self.factory.client.transport.loseConnection()
 
     @defer.inlineCallbacks
     def test_reconnect(self):
@@ -1643,6 +1784,7 @@ class TestFactory(unittest.TestCase):
 
         a = yield self.factory.client.info()
         self.assertTrue('uptime_in_days' in a)
+    timeout = 4
 
 
 class ProtocolBuffering(Protocol):
@@ -1691,7 +1833,6 @@ class PubSub(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_subscribe(self):
-        r = self.redis
         s = self.subscriber
         t = self.assertEqual
 
@@ -1713,9 +1854,7 @@ class PubSub(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_unsubscribe(self):
-        r = self.redis
         s = self.subscriber
-        t = self.assertEqual
 
         cb = s.channel_subscribed
         yield s.subscribe("channelA", "channelB", "channelC")
@@ -1729,7 +1868,6 @@ class PubSub(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_psubscribe(self):
-        r = self.redis
         s = self.subscriber
         t = self.assertEqual
 
@@ -1751,9 +1889,7 @@ class PubSub(CommandsTestBase):
 
     @defer.inlineCallbacks
     def test_punsubscribe(self):
-        r = self.redis
         s = self.subscriber
-        t = self.assertEqual
 
         cb = s.channel_subscribed
         yield s.psubscribe("channel*", "magic*", "woot*")
